//@author: schlonzo

Texture2D texture2d <string uiname="Texture";>;

SamplerState sL : IMMUTABLE
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
};
 
cbuffer cbPerDraw : register( b0 )
{
	float4x4 tVP : LAYERVIEWPROJECTION;	
};

cbuffer cbPerObj : register( b1 )
{
	float4x4 tW : WORLD;
	float4x4 tWV : WORDVIEWPROJECTION;
	float4x4 tWIT : WORLDINVERSETRANSPOSE;
	float4 cAmb <bool color=true;String uiname="Color";> = { 1.0f,1.0f,1.0f,1.0f };
	
	float edgeWidth;
	float pctExtend;
};

struct vsIn
{
	//model must provide normals!
	float4 PosO : POSITION;	
	float4 PosN : NORMAL;
	float4 TexCd : TEXCOORD0;
};

// registers must match when piping through
// this means position of variable in struct
// VS[0] -> GS [0] -> PS [0]

struct gsIn
{
    float4 PosWVP: SV_Position;
	float4 Normal : NORMAL; 
    float4 TexCd: TEXCOORD0;
	float4 PosW: POSITION;
};

struct psIn
{
    float4 PosWVP: SV_Position;
	float4 Normal : NORMAL; 
    float4 TexCd: TEXCOORD0;
	bool isEdge: POINT;
};

gsIn VS(vsIn input)
{
    gsIn output;	
	
	
	//Normal in tangent space
	float3x3 normalMatrix = (float3x3)tWIT; // world inverse transpose!
	float3 vertexNormal = normalize(mul(input.PosN.xyz, normalMatrix)); //transform to tangent space
	output.Normal = float4(vertexNormal,0);
	
	float4 PosW  = mul(input.PosO,tW);	 
    output.PosWVP  = mul(PosW,tVP);
	output.PosW = PosW;
    output.TexCd = input.TexCd;
    return output;
}

// is Z Coordinate in Tangent space facing towards camera ( Z > 0 )
bool isFrontFacing(float3 a, float3 b, float3 c)
{
	return(	(a.x * b.y - b.x * a.y) +
			(b.x * c.y - c.x * b.y) +
			(c.x * a.y - a.x * c.y) > 0 );
}

// triangle stream must be inout parameter?
void emitQuad(float3 e0, float3 e1, inout TriangleStream<psIn> tStream)
{
	float2 ext = pctExtend * (e1.xy - e0.xy);
	float2 v = normalize(e1.xy- e0.xy);
	float2 n = float2(v.y, -v.x) * edgeWidth;
	
	psIn output = (psIn)0;
	
	// triangles are edges
	output.isEdge = true;
	
	//emitt quad vertices
	output.PosWVP = float4(e0.xy-ext, e0.z, 1);
	tStream.Append(output);
	output.PosWVP = float4(e0.xy-n-ext, e0.z, 1);
	tStream.Append(output);
	output.PosWVP = float4(e1.xy+ext, e1.z, 1);
	tStream.Append(output);
	output.PosWVP = float4(e1.xy-n+ext, e1.z, 1);
	tStream.Append(output);
	
	//restart so triangles won't stick together
	tStream.RestartStrip();
}

//triangleadj = 6 Vertices [V0, A0, V1, A1, V2, A2]
// V = Vertex main triangle, A = Adjacent vertex 
//maxvertex = orinal triangle (3) + quads on each corner possible (3*3*2=12)= 15!!

[maxvertexcount(15)]
void GS( triangleadj gsIn input[6], inout TriangleStream<psIn> tStream )
{  
	psIn output = (psIn)0;	
	
	//convert from homogenous representation to true cartesian value
	float3 p[6];
	for(uint i =0; i<6; i++)
	{
		p[i] = input[i].PosWVP.xyz / input[i].PosWVP.w;
	}
	
	
	//check if main triangle faces
	if( isFrontFacing(p[0], p[2], p[4]) )
	{
		//then check if the adjacent ones also do
		//and emit quad(s) on adjacent line(s) of triangle
		if( ! isFrontFacing(p[0], p[1], p[2]) ) emitQuad(p[0], p[2], tStream); 	
		if( ! isFrontFacing(p[2], p[3], p[4]) ) emitQuad(p[2], p[4], tStream); 	
		if( ! isFrontFacing(p[4], p[5], p[0]) ) emitQuad(p[4], p[0], tStream);
	}
	
	// triangles are original model
	output.isEdge = false;
	
	//emit original triangle
	//mind CW vs CCW = front vs. backfacing triangle
	uint3 mainTriangle = uint3(0,4,2); 
	for(uint j=0; j<3; j++)
	{
		output.PosWVP = input[ mainTriangle[j]].PosWVP;
		tStream.Append(output);
	}
	
	//restart so triangles won't stick together
	tStream.RestartStrip();	
}


float4 PS(psIn In): SV_Target
{
	float4 col = float4(0,0,0,1);
	
	if(In.isEdge)
	{
		col = cAmb;
	}else{
		col = texture2d.Sample(sL,In.TexCd.xy);
	}	

    return col;
}

technique10 Constant
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_5_0, VS() ) );
		SetGeometryShader( CompileShader( gs_5_0, GS() ) );
		SetPixelShader( CompileShader( ps_5_0, PS() ) );
	}
}




